/* Copyright (C) 2000 Michael Elkins <me@mutt.org>
 * This is free software distributed under the terms of the GNU Public License.
 * See the file COPYING for details.
 */

/* A very simple SMTP relay client */

#include <sysexits.h>
#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <pwd.h>

int Verbose = 0;

void
usage (void)
{
  fprintf (stderr,
	   "usage: %s [ -hvV ] [ -f FROM ] [ -S SMTPSERVER ] [ -p PORT ] RECIPIENT ...\n",
	   PACKAGE);
  fputs ("  -f FROM	use FROM as envelope return address\n", stderr);
  fputs ("  -h		print this help message\n", stderr);
  fputs ("  -p PORT	connect on PORT (default: 25)\n", stderr);
  fputs ("  -S SERVER	use SERVER as SMTP relay (default: localhost)\n",
	 stderr);
  fputs ("  -v		display version number\n", stderr);
  fputs ("  -V		turn on verbose mode", stderr);
  exit (EX_USAGE);
}

void
version (void)
{
  printf ("%s %s Copyright (C) 2000-1 Michael Elkins <me@mutt.org>\n",
	  PACKAGE, VERSION);
  puts ("http://www.sigpipe.org/smtppush/");
  exit (EX_OK);
}

int
get_server_message (int fd)
{
  char buf[1024];
  int buflen = 0;
  int n;

  while ((unsigned) buflen < sizeof (buf))
    {
      n = read (fd, &buf[buflen], 1);
      if (n <= 0)
	return EX_IOERR;
      if (buf[buflen] == '\r')
	{
	  n = read (fd, &buf[buflen], 1);
	  if (n <= 0)
	    return EX_IOERR;
	}
      if (buf[buflen] == '\n')
	{
	  buf[buflen] = 0;
	  n = strtol (buf, NULL, 10);
	  if (Verbose || n >= 400)
	    {
	      /* some error occurred */
	      fprintf (stderr, "%s: %s\n", PACKAGE, buf);
	      if (n >= 400)
		return EX_PROTOCOL;
	    }
	  return 0;
	}
      buflen++;
    }
  /* line was too long */
  return EX_PROTOCOL;
}

int
send_to_server (int fd, const char *fmt, ...)
{
  va_list ap;
  char buf[1024];

  va_start (ap, fmt);
  vsnprintf (buf, sizeof (buf) - 3, fmt, ap);
  va_end (ap);

  if (Verbose)
    fprintf (stderr, "%s: <<< %s\n", PACKAGE, buf);

  /* safe since we only used sizeof(buf)-3 above */
  strcat (buf, "\r\n");
  if (write (fd, buf, strlen (buf)) != (int) strlen (buf))
    return -1;
  return 0;
}

int
main (int argc, char **argv)
{
  int i;
  char *from = NULL;
  char *host = "localhost";
  int port = 25;
  struct sockaddr_in sin;
  struct hostent *he;
  int sock;
  char *myhost = NULL;
  char buf[1024];
  char realhost[256];
  char realfrom[256];

  while ((i = getopt (argc, argv, "hvf:S:p:V")) != -1)
    {
      switch (i)
	{
	case 'f':
	  from = optarg;
	  break;
	case 'S':
	  host = optarg;
	  break;
	case 'p':
	  port = atoi (optarg);
	  break;
	case 'V':
	  Verbose = 1;
	  break;
	case 'v':
	  version ();
	case 'h':
	default:
	  usage ();
	}
    }
  if (argv[optind] == NULL)
    usage ();

  /* if the return envelope address was set, use the host part as this
   * hosts dns name
   */
  if (from)
    {
      myhost = strchr (from, '@');
      if (!myhost)
	{
	  fprintf (stderr, "%s: no @ found in return address\n", PACKAGE);
	  return EX_USAGE;
	}
      myhost++;
    }
  else
    {
      struct passwd *pw;

      /* otherwise we need to look it up */
      if (gethostname (realhost, sizeof (realhost)))
	{
	  fprintf (stderr, "%s: gethostname: %s (errno %d)\n", PACKAGE,
		   strerror (errno), errno);
	  return EX_OSERR;
	}
      if (Verbose)
	fprintf (stderr, "%s: local host name is %s\n", PACKAGE, realhost);

      /* look up the fqdn */
      he = gethostbyname (realhost);
      if (!he)
	{
	  fprintf (stderr, "%s: gethostbyname: %s: %s (errno %d)\n",
		   PACKAGE, realhost, strerror (h_errno), h_errno);
	  exit (EX_NOHOST);
	}
      strcpy (realhost, he->h_name);
      if (Verbose)
	fprintf (stderr, "%s: fqdn is %s\n", PACKAGE, realhost);

      /* find the user's login name */
      pw = getpwuid (getuid ());
      if (!pw)
	{
	  fprintf (stderr, "%s: getpwuid: %s (errno %d)\n", PACKAGE,
		   strerror (errno), errno);
	  return EX_OSERR;
	}

      /* form the return address */
      snprintf (realfrom, sizeof (realfrom), "%s@%s", pw->pw_name, realhost);
      if (Verbose)
	fprintf (stderr, "%s: realfrom is %s\n", PACKAGE, realfrom);

      from = realfrom;
      myhost = realhost;
    }

  memset (&sin, 0, sizeof (sin));
  sin.sin_family = AF_INET;
  sin.sin_port = htons (port);
  if (inet_aton (host, &sin.sin_addr) == 0)
    {
      /* not an ip address, look up hostname */
      he = gethostbyname (host);
      if (!he)
	{
	  fprintf (stderr, "%s: gethostbyname: %s: %s (errno %d)\n",
		   PACKAGE, host, strerror (h_errno), h_errno);
	  exit (EX_NOHOST);
	}
      sin.sin_addr.s_addr = *((int *) he->h_addr_list[0]);
    }

  sock = socket (AF_INET, SOCK_STREAM, 0);
  if (sock < 0)
    {
      fprintf (stderr, "%s: socket: %s (errno %d)\n", PACKAGE,
	       strerror (errno), errno);
      exit (EX_OSERR);
    }

  if (connect (sock, (struct sockaddr *) &sin, sizeof (sin)))
    {
      fprintf (stderr, "%s: connect: %s (errno %d)\n", PACKAGE,
	       strerror (errno), errno);
      exit (EX_OSERR);
    }

  /* eat the server's greeting */
  i = get_server_message (sock);
  if (i != EX_OK)
    exit (i);

  if (send_to_server (sock, "HELO %s", myhost))
    exit (EX_IOERR);
  i = get_server_message (sock);
  if (i != EX_OK)
    exit (i);

  if (send_to_server (sock, "MAIL FROM: <%s>", from))
    exit (EX_IOERR);
  i = get_server_message (sock);
  if (i != EX_OK)
    exit (i);

  while (argv[optind])
    {
      if (send_to_server (sock, "RCPT TO: <%s>", argv[optind]))
	exit (EX_IOERR);
      i = get_server_message (sock);
      if (i != EX_OK)
	exit (i);
      optind++;
    }

  if (send_to_server (sock, "DATA", from))
    exit (EX_IOERR);
  i = get_server_message (sock);
  if (i != EX_OK)
    exit (i);

  /* copy the message from stdin */
  while (fgets (buf, sizeof (buf) - 1, stdin))
    {
      buf[strlen (buf) - 1] = 0;
      if (buf[0] == '.')
	if (write (sock, ".", 1) != 1)
	  exit (EX_IOERR);
      if (write (sock, buf, strlen (buf)) != (int) strlen (buf))
	exit (EX_IOERR);
      if (write (sock, "\r\n", 2) != 2)
	exit (EX_IOERR);
    }

  if (send_to_server (sock, ".", from))
    exit (EX_IOERR);
  i = get_server_message (sock);
  if (i != EX_OK)
    exit (i);

  close (sock);

  exit (EX_OK);
}
